"""
This is a fix since currently Lctakes does not deal with concept-internal
negation like: 'opioids no abuse'; currently, this will resolve to a positive
mention.

"""

import argparse
import copy
import logging
import logging.config
import pyodbc

from jinja2 import Template

from pytakes import templates, header
from .util import mylogger
from .util.utils import get_valid_args

from .nlp.statusminer import MyStatusTagger, sort_rules_for_status, get_context, StatusMiner
from .util.db_reader import DbInterface


def lists_are_equal(lst1, lst2):
    for el1, el2 in zip(lst1, lst2):
        if el1 != el2:
            return False
    return True


def insert_into(dbi, table, cols, row):
    """
    :param dbi:
    :param table:
    :param cols:
    :param row:

    """
    sql = Template(templates.PP_INSERT_INTO).render({
        'table_name': table,
        'labels': cols[1:],
        'values': row[1:]
    })
    try:
        dbi.execute_commit(sql)
    except Exception as e:
        logging.error(sql)
        logging.exception(e)
        raise e


def get_input_data(dbi, input_table, columns):
    sql = Template(templates.PP_INPUT_DATA).render({
        'input_table': input_table,
        'labels': columns
    })
    return dbi.execute_fetchall(sql)


def create_destination_table(dbi, input_table, output_table):
    """
    Copy all attributes of input to output table.
    :param dbi:
    :param input_table:
    :param output_table:
    """
    sql = Template(templates.PP_DEST_TABLE).render({
        'dest_table': output_table,
        'input_table': input_table
    })
    try:
        dbi.execute_commit(sql)
        prepare_output_tables(dbi, output_table)
    except pyodbc.ProgrammingError as pe:
        logging.warning('Table "%s" already exists.' % output_table)
        return


def prepare_output_tables(dbi, output_table):
    """

    :param dbi:
    :param output_table:
    :return:
    """
    sql = Template(templates.PP_PREP_DEST_TABLE).render({
        'table_name': output_table
    })
    dbi.execute_commit(sql)


def postprocess(dbi,
                negation_table,
                negation_variation,
                input_table,
                input_column,
                output_table,
                batch_count,
                tracking_method):
    """
    :param tracking_method:
    :param batch_count:
    :param dbi:
    :param negation_table:
    :param negation_variation:
    :param input_table:
    :param input_column:
    :param output_table:

    """
    tagger = MyStatusTagger(sort_rules_for_status(get_context(dbi, negation_table)), rx_var=negation_variation)

    if tracking_method == 'name':
        first_input_table = '{}_{}'.format(input_table, 1)  # input table for first instance
    elif tracking_method == 'column':
        first_input_table = input_table
    else:
        raise ValueError('Unrecognized tracking method: "{}".'.format(tracking_method))

    columns = dbi.get_table_columns(first_input_table)
    out_columns = list(columns)
    out_columns.append('updated')
    create_destination_table(dbi, first_input_table, output_table)
    if input_column in out_columns:
        col_idx = out_columns.index(input_column)
    else:
        raise ValueError('Unrecognized column "%s" in table %s.' % (input_column, input_table))

    for i in range(1, batch_count):
        if tracking_method == 'name':
            curr_input_table = '{}_{}'.format(input_table, 1)  # input table for first instance
        elif tracking_method == 'column':
            curr_input_table = input_table
        else:
            raise ValueError('Unrecognized tracking method: "{}".'.format(tracking_method))

        data = get_input_data(dbi, curr_input_table, columns)
        for row in data:
            new_row = []
            for el in row:
                if isinstance(el, int):
                    new_row.append(el)
                else:
                    new_row.append(el.encode('utf-8').decode('utf-8', 'ignore'))
            row = new_row
            text = row[col_idx]
            orig_row = copy.copy(row)
            for negConcept in tagger.find_negation(text):
                _type = negConcept.type().lower()
                if _type == header.NEGATION_TYPE:
                    row[out_columns.index(header.CERTAINTY)] = 0
                elif _type == header.IMPROBABLE_TYPE:
                    row[out_columns.index(header.CERTAINTY)] = 1
                elif _type == header.POSSIBLE_TYPE:
                    row[out_columns.index(header.CERTAINTY)] = 2
                elif _type == header.PROBABLE_TYPE:
                    row[out_columns.index(header.CERTAINTY)] = 3
                if _type == header.HYPOTHETICAL_TYPE:
                    row[out_columns.index(header.HYPOTHETICAL)] = 1
                if _type == header.FUTURE_TYPE:
                    row[out_columns.index(header.HYPOTHETICAL)] = 1
                if _type == header.HISTORICAL_TYPE:
                    row[out_columns.index(header.HISTORICAL)] = 1
                if _type == header.OTHER_SUBJECT_TYPE:
                    row[out_columns.index(header.OTHER_SUBJECT)] = 1

            if lists_are_equal(orig_row, row):
                row.append(0)
            else:
                row.append(1)

            insert_into(dbi, output_table, out_columns, row)

    return True


def main():
    parser = argparse.ArgumentParser(fromfile_prefix_chars='@')
    parser.add_argument('--driver', required=False, default='SQL Server')
    parser.add_argument('-s', '--server', required=True, help='Database server to use.')
    parser.add_argument('-d', '--database', required=True, help='Database to use.')
    parser.add_argument('--negation-table', help='Table of negation triggers, along with role.')
    parser.add_argument('--negation-variation', type=int, default=0, required=False,
                        help='Amount of variation to allow in negations. Values: 0-3.')
    parser.add_argument('--input-table', help='Table that has a column which needs to be evaluated by negex.')
    parser.add_argument('--input-column', help='Column name which needs to be modified.')
    parser.add_argument('--output-table', help='Output table.')
    parser.add_argument('--batch-count', type=int, help='Batch count.')
    parser.add_argument('--tracking-method', choices=['name', 'column'], default='name',
                        help='Method to track progress of batches.')

    parser.add_argument('--verbosity', type=int, default=2, help='Verbosity of log output.')
    args = parser.parse_args()

    loglevel = mylogger.resolve_verbosity(args.verbosity)
    logging.config.dictConfig(mylogger.setup(name='postprocessor', loglevel=loglevel))

    dbi = DbInterface(args.driver, args.server, args.database, loglevel)

    try:
        postprocess(dbi, **get_valid_args(postprocess, vars(args)))
    except Exception as e:
        logging.exception(e)
        logging.error('Process terminated.')


if __name__ == '__main__':
    main()



"""Removing file content from the logic of automate run.

"""

# automate run details
RUN_BATCH_FILE = r'''@echo off
echo Running batch {{ batch_number }}.
{{ python }} {{ pytakes_path }}processor.py "@.\pytakes-batch{{ batch_number }}.conf"
if %%errorlevel%% equ 0 (
{% if send_email %}
{{ python }} {{ pytakes_path }}sendmail.py -s "Batch {{ batch_number }} Completed" "@.\email.conf"
{% endif %}
echo Successful.
) else (
{% if send_email %}
{{ python }} {{ pytakes_path }}sendmail.py -s "Batch {{ batch_number }} Failed: Log Included" -f ".\log\pytakes-processor{{ batch_number }}.log" "@.\bad_email.conf"
{% endif %}
echo Failed.
)
pause
'''

RUN_COMMAND_BATCH_FILE = r'''@echo off
echo Running batch {{ batch_number }}.
pytakes-processor "@.\pytakes-batch{{ batch_number }}.conf"
if %%errorlevel%% equ 0 (
{% if send_email %}
pytakes-sendmail -s "Batch {{ batch_number }} Completed" "@.\email.conf"
{% endif %}
echo Successful.
) else (
{% if send_email %}
pytakes-sendmail -s "Batch {{ batch_number }} Failed: Log Included" -f ".\log\pytakes-processor{{ batch_number }}.log" "@.\bad_email.conf"
{% endif %}
echo Failed.
)
pause
'''

RUN_CONF_FILE = r'''--driver={{ driver }}
--server={{ server }}
--database={{ database }}
--document-table={{ document_table }}
--meta-labels
{{ primary_key }}
{%- for meta_label in meta_labels %}
{{ meta_label }}
{%- endfor %}
--text-labels
{%- for text_label in text_labels %}
{{ text_label }}
{%- endfor %}
--tracking-method={{ tracking_method }}
--destination-table={{ destination_table }}_pre
{%- for option in options %}
{{ option }}
{%- endfor %}
--batch-mode={{ primary_key }}
--batch-size={{ batch_size }}
--batch-number
{{ batch_start }}
{{ batch_end }}
'''

EMAIL_CONF_FILE = r'''{% for recipient in recipients %}--recipient
{{ recipient }}
{%- endfor %}
--server-address={{ mail_server_address }}
--sender
{{ sender }}
--text
This notification is to inform you that another batch ({{ filecount }} total) has been completed for table {{ destination_table }}.
'''

BAD_EMAIL_CONF_FILE = r'''{%- for recipient in recipients %}--recipient
{{ recipient }}
{%- endfor %}
--server-address={{ mail_server_address }}
--sender
{{ sender }}
--text
This notification is to inform you that a batch ({{ filecount }} total) has failed for table {{ destination_table }}.

The log file is attached for debugging.
'''

PP_BATCH_FILE = r'''{{ python }} {{ pytakes_path }}postprocessor.py "@.\postprocess.conf"
pause
'''

PP_COMMAND_BATCH_FILE = r'''pytakes-postprocessor "@.\postprocess.conf"
pause
'''

PP_CONF_FILE = r'''--driver={{ driver }}
--server={{ server }}
--database={{ database }}
--input-table={{ destination_table }}_pre
--output-table={{ destination_table }}
{%- if negation_table and negation_variation %}
--negation-table={{ negation_table }}
--negation-variation={{ negation_variation }}
{%- endif %}
--input-column=captured
--batch-count={{ batch_count }}
--tracking-method={{ tracking_method }}
'''

SAMPLE_CONF_FILE = r'''--driver=DRIVER
--server=SERVER
--database=DB
--dictionary-table=DICTIONARY_TABLE
--negation-table=NEGATION_TABLE
--negation-variation=0
--document-table=DOCUMENT_TABLE
--output-dir=DIRECTORY
--destination-table=DESTINATION_TABLE
--max-intervening-terms=0
--max-length-of-search=1
--meta-labels
doc_id
pat_id
date
--primary-key
doc_id
--text-labels
note_text
--mail-server-address
mail.test.org
--sender
Automated Email,example@example.com
--recipients
Recipient Name,example@example.com
Recipient2 Name,example2@example.com
'''

# for processor.py
PROC_INSERT_INTO2_QUERY = r'''INSERT INTO {{ destination_table }} (
{%- for label in labels %}
{{ label }}{% if not loop.last %},{% endif %}
{%- endfor %}
) VALUES (
{%- for meta in metas %}
'{{ meta|replace("'", "''") }}',
{%- endfor %}
{{ feature.id() }},
'{{ captured|replace("'", "''") }}',
'{{ context|replace("'", "''") }}',
'{{ text|replace("'", "''") }}',
{{ feature.get_certainty() }},
{% if feature.is_hypothetical() %}1{% else %}0{% endif %},
{% if feature.is_historical() %}1{% else %}0{% endif %},
{% if feature.is_hypothetical() %}1{% else %}0{% endif %},
{{ feature.begin() }},
{{ feature.end() }},
{{ feature.get_absolute_begin() }},
{{ feature.get_absolute_end() }}
{% if hostname %}, '{{ hostname|replace("'", "''") }}', {{ batch_number }}{% endif %}
)
'''

PROC_INSERT_INTO3_QUERY = r'''INSERT INTO {{ destination_table }} (
{%- for label in labels %}
{{ label }}{% if not loop.last %},{% endif %}
{%- endfor %}
) VALUES (
{%- for meta in metas %}
'{{ meta }}',
{%- endfor %}
{{ feature.get_id() }},
{{ feature.get_feature() }},
{{ feature.get_category() }}
{% if hostname %}, {{ hostname }}, {{ batch_number }}{% endif %}
)
'''

PROC_GET_DOC_IDS = r'''SELECT {{ table_id }}
FROM {{ doc_table }}
{%- if order_by %}
ORDER BY {{ order_by }}
{%- endif %}
'''

PROC_CREATE_TABLE = r'''CREATE TABLE {{ destination_table }} (
rowid int IDENTITY(1, 1) PRIMARY KEY,
{%- for label, column_type in labels_types %}
{{ label }} {{ column_type }}{% if not loop.last %},{% endif %}
{%- endfor %}
)
'''

PROC_GET_CONTEXT = r'''SELECT negex, type, direction
FROM {{ neg_table }}
'''

PROC_GET_TERMS = r'''SELECT id
    , text
    , cui
    ,{% if 'valence' not in columns %}  {% if valence %} {{ valence }} {% else %} 1 {% endif %}  as {% endif %} valence
    ,{% if 'regexvariation' not in columns %} {% if regex_variation %} {{ regex_variation }} {% else %} 3 {% endif %} as {% endif %} regexvariation
    ,{% if 'WordOrder' not in columns %} {% if word_order %} {{ word_order }} {% else %} 1 {% endif %} as {% endif %}  wordorder
FROM {{ term_table }}
'''

PROC_GET_DOCS = r'''SELECT
{%- if where_clause and order_by %}
TOP {{ batch_size }}
{%- endif %}
{%- for meta in meta_labels %}
{{ meta }},
{%- endfor %}
{%- for text in text_labels %}
{{ text }}{% if not loop.last %},{% endif %}
{%- endfor %}
FROM {{ doc_table }}
{%- if where_clause and order_by %}
WHERE {{ where_clause }}
ORDER BY {{ order_by }}
{%- endif %}
'''

# for postprocessor.py
PP_INSERT_INTO = r'''INSERT INTO {{ table_name }} (
{%- for label in labels %}
{{ label }} {% if not loop.last %},{% endif %}
{%- endfor %}
) VALUES (
{%- for value in values %}
    {%- if value is number %}
    {{ value }}
    {%- else %}
    '{{ value|replace("'", "''") }}'
    {%- endif %}
     {% if not loop.last %},{% endif %}
{%- endfor %}
)
'''

PP_INPUT_DATA = r'''SELECT
{%- for label in labels %}
{{ label }} {% if not loop.last %},{% endif %}
{%- endfor %}
FROM {{ input_table }}
'''

PP_PREP_DEST_TABLE = r'''ALTER TABLE {{ table_name }}
ADD updated int
'''

PP_DEST_TABLE = r'''SELECT *
INTO {{ dest_table }}
FROM {{ input_table }}
WHERE 1 = 2
'''